searchNodes=[{"type":"module","title":"Tortoise311","doc":"A MQTT client for Elixir. Tortoise311 provides ways of publishing messages to, and receiving messages from one or many MQTT brokers via TCP or SSL. The design philosophy of Tortoise311 is to hide the protocol specific details from the user, and expose interfaces and a connection life cycle that should feel natural to Elixir, while not limiting the capability of what one can do with the MQTT protocol. First off, connection to a broker happens through a connection specification. This results in a process that can be supervised, either by the application the connection should live and die with, or by being supervised by the Tortoise311 application itself. Once the connection is established the Tortoise311 application should do its best to keep that connection open, by automatically sending keep alive messages (as the protocol specifies), and eventually attempt to reconnect if the connection should drop. Secondly, a connection is specified with a user defined callback module, following the Tortoise311.Handler -behaviour, which allow the user to hook into certain events happening in the life cycle of the connection. This way code can get executed when: The connection is established The client has been disconnected from the broker A topic filter subscription has been accepted (or declined) A topic filter has been successfully unsubscribed A message is received on one of the subscribed topic filters Besides this there are hooks for the usual life-cycle events one would expect, such as init/1 and terminate/2 . Thirdly, publishing is handled in such a way that the semantics of the levels of Quality of Service, specified by the MQTT protocol, is mapped to the Elixir message passing semantics. Tortoise311 expose an interface for publishing messages that hide the protocol details of message delivery (retrieval of acknowledge, release, complete messages) and instead provide Tortoise311.publish/4 which will deliver the message to the broker and receive a response in the process mailbox when a message with a QoS&gt;0 has been handed to the server. This allow the user to keep track of the messages that has been delivered, or simply by using the Tortoise311.publish_sync/4 form that will block the calling process until the message has been safely handed to the broker. Messages with QoS1 or QoS2 are stored in a process until they are delivered, so once they are published the client should retry delivery to make sure they reach their destination. An alternative way of posting messages is implemented in Tortoise311.Pipe , which provide a data structure that among other things keep a reference to the connection socket. This allow for an efficient way of posting messages because the data can get shot directly onto the wire without having to copy the message between processes (unless the message has a QoS of 1 or 2, in which case they will end up in a process to ensure they will get delivered). The pipe will automatically renew its connection socket if the connection has been dropped, so ideally this message sending approach should be fast and efficient.","ref":"Tortoise311.html"},{"type":"function","title":"Tortoise311.default_timeout/0","doc":"The default timeout value","ref":"Tortoise311.html#default_timeout/0"},{"type":"function","title":"Tortoise311.publish/4","doc":"Publish a message to the MQTT broker. The publish function requires a client_id and a valid MQTT topic. If no payload is set an empty zero byte message will get send to the broker. Optionally an options list can get passed to the publish, making it possible to specify if the message should be retained on the server, and with what quality of service the message should be published with. retain indicates, when set to true , that the broker should retain the message for the topic. Retained messages are delivered to clients when they subscribe to the topic. Only one message at a time can be retained for a given topic, so sending a new one will overwrite the old. retain defaults to false . qos set the quality of service, and integer of 0, 1, or 2. The qos defaults to 0 . Publishing a message with the payload hello to to topic foo/bar with a QoS1 could look like this: Tortoise311 . publish ( &quot;client_id&quot; , &quot;foo/bar&quot; , &quot;hello&quot; , qos : 1 ) Notice that if you want to send a message with an empty payload with options you will have to set to payload to nil like this: Tortoise311 . publish ( &quot;client_id&quot; , &quot;foo/bar&quot; , nil , retain : true ) Return Values The specified Quality of Service for a given publish will alter the behaviour of the return value. When publishing a message with a QoS0 an :ok will simply get returned. This is because a QoS0 is a &quot;fire and forget.&quot; There are no quality of service so no efforts are made to ensure that the message will reach its destination (though it very likely will). :ok = Tortoise311 . publish ( &quot;client_id&quot; , &quot;foo/bar&quot; , nil , qos : 0 ) When a message is published using either a QoS1 or QoS2, Tortoise311 will ensure that the message is delivered. A unique reference will get returned and eventually a message will get delivered to the process mailbox, containing the result of the publish when it has been handed over: { :ok , ref } = Tortoise311 . publish ( &quot;client_id&quot; , &quot;foo/bar&quot; , nil , qos : 2 ) receive do { { Tortoise311 , &quot;client_id&quot; } , ^ ref , result } -&gt; IO . inspect ( { :result , result } ) after 5000 -&gt; { :error , :timeout } end Be sure to implement a handle_info/2 in GenServer processes that publish messages using Tortoise311.publish/4. Notice that the returned message has a structure: { { Tortoise311 , &quot;client_id&quot; } , ^ ref , result } It is possible to send to multiple clients and blanket match on results designated for a given client id, and the message is tagged with Tortoise311 so it is easy to see where the message originated from.","ref":"Tortoise311.html#publish/4"},{"type":"function","title":"Tortoise311.publish_sync/4","doc":"Synchronously send a message to the MQTT broker. This is very similar to Tortoise311.publish/4 with the difference that it will block the calling process until the message has been handed over to the server; the configuration options are the same with the addition of the timeout option which specifies how long we are willing to wait for a reply. Per default the timeout is set to Tortoise311.default_timeout() , it is advisable to set it to a reasonable amount in milliseconds as it otherwise could block forever. msg = &quot;Hello, from the World of Tomorrow !&quot; case Tortoise311 . publish_sync ( &quot;my_client_id&quot; , &quot;foo/bar&quot; , msg , qos : 2 , timeout : 200 ) do :ok -&gt; :done { :error , :timeout } -&gt; :timeout end Notice: It does not make sense to use publish_sync/4 on a publish that has a QoS=0, because that will return instantly anyways. It is made possible for consistency, and it is the default QoS. See the documentation for Tortoise311.publish/4 for configuration.","ref":"Tortoise311.html#publish_sync/4"},{"type":"type","title":"Tortoise311.client_id/0","doc":"An identifier used to identify the client on the server. Most servers accept a maximum of 23 UTF-8 encode bytes for a client id, and only the characters: &quot;0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ&quot; Tortoise311 accept atoms as client ids but they it will be converted to a string before going on the wire. Be careful with atoms such as Example because they are expanded to the atom :&quot;Elixir.Example&quot; , it is really easy to hit the maximum byte limit. Solving this is easy, just add a : before the client id such as :Example .","ref":"Tortoise311.html#t:client_id/0"},{"type":"type","title":"Tortoise311.package_identifier/0","doc":"A 16-bit number identifying a message in a message exchange. Some MQTT packages are part of a message exchange and need an identifier so the server and client can distinct between multiple in-flight messages. Tortoise311 will assign package identifier to packages that need them, so outside of tests (where it is beneficial to assert on the identifier of a package) it should be set by tortoise itself; so just leave it as nil .","ref":"Tortoise311.html#t:package_identifier/0"},{"type":"type","title":"Tortoise311.payload/0","doc":"An optional message payload. A message can optionally have a payload. The payload is a series of bytes and for MQTT 3.1.1 the payload has no defined structure; any series of bytes will do, and the client has to make sense of it. The payload will be nil if there is no payload. This is done to distinct between a zero byte binary and an empty payload.","ref":"Tortoise311.html#t:payload/0"},{"type":"type","title":"Tortoise311.qos/0","doc":"What Quality of Service (QoS) mode should be used. Quality of Service is one of 0, 1, and 2 denoting the following: 0 no quality of service. The message is a fire and forget. 1 at least once delivery. The receiver will respond with an acknowledge message, so the sender will be certain that the message has reached the destination. It is possible that a message will be delivered twice though, as the package identifier for a publish will be relinquished when the message has been acknowledged, so a package with the same identifier will be treated as a new message though it might be a re-transmission. 2 exactly once delivery. The receiver will only receive the message once. This happens by having a more elaborate message exchange than the QoS=1 variant. There are a difference in the semantics of assigning a QoS to a publish and a subscription. When assigned to a publish the message will get delivered to the server with the requested QoS; that is if it accept that level of QoS for the given topic. When used in the context of a subscription it should be read as the maximum QoS . When messages are published to the subscribed topic the message will get on-warded with the same topic as it was delivered with, or downgraded to the maximum QoS of the subscription for the given subscribing client. That is, if the client subscribe with a maximum QoS=2 and a message is published to said topic with a QoS=1, the message will get downgraded to QoS=1 when on-warded to the client.","ref":"Tortoise311.html#t:qos/0"},{"type":"type","title":"Tortoise311.topic/0","doc":"A topic for a message. According to the MQTT 3.1.1 specification a valid topic must be at least one character long. They are case sensitive and can include space characters. MQTT topics consist of topic levels which are delimited with forward slashes / . A topic with a leading or trailing forward slash is allowed but they create distinct topics from the ones without; /sports/tennis/results are different from sports/tennis/results . While a topic level normally require at least one character the topic / (a single forward slash) is valid. The server will drop the connection if it receive an invalid topic.","ref":"Tortoise311.html#t:topic/0"},{"type":"type","title":"Tortoise311.topic_filter/0","doc":"A topic filter for a subscription. The topic filter is different from a topic because it is allowed to contain wildcard characters: + is a single level wildcard which is allowed to stand on any position in the topic filter. For instance: sport/+/results will match sport/tennis/results , sport/soccer/results , etc. # is a multi-level wildcard and is only allowed to be on the last position of the topic filter. For instance: sport/# will match sport/tennis/results , sport/tennis/announcements , etc. The server will reject any invalid topic filter and close the connection.","ref":"Tortoise311.html#t:topic_filter/0"},{"type":"module","title":"Tortoise311.Connection","doc":"Establish a connection to a MQTT broker. Todo.","ref":"Tortoise311.Connection.html"},{"type":"function","title":"Tortoise311.Connection.child_spec/1","doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Tortoise311.Connection.html#child_spec/1"},{"type":"function","title":"Tortoise311.Connection.disconnect/1","doc":"Close the connection to the broker. Given the client_id of a running connection it will cancel the inflight messages and send the proper disconnect message to the broker. The session will get terminated on the server.","ref":"Tortoise311.Connection.html#disconnect/1"},{"type":"function","title":"Tortoise311.Connection.ping/1","doc":"Ping the broker. When the round-trip is complete a message with the time taken in milliseconds will be send to the process that invoked the ping command. The connection will automatically ping the broker at the interval specified in the connection configuration, so there is no need to setup a reoccurring ping. This ping function is exposed for debugging purposes. If ping latency over time is desired it is better to listen on :ping_response using the Tortoise311.Events PubSub.","ref":"Tortoise311.Connection.html#ping/1"},{"type":"function","title":"Tortoise311.Connection.ping_sync/2","doc":"Ping the server and await the ping latency reply. Takes a client_id and an optional timeout . Like ping/1 but will block the caller process until a response is received from the server. The response will contain the ping latency in milliseconds.","ref":"Tortoise311.Connection.html#ping_sync/2"},{"type":"function","title":"Tortoise311.Connection.start_link/2","doc":"Start a connection process and link it to the current process. Read the documentation on child_spec/1 if you want... (todo!)","ref":"Tortoise311.Connection.html#start_link/2"},{"type":"function","title":"Tortoise311.Connection.subscribe/3","doc":"Subscribe to one or more topics using topic filters on client_id The topic filter should be a 2-tuple, {topic_filter, qos} , where the topic_filter is a valid MQTT topic filter, and qos an integer value 0 through 2. Multiple topics can be given as a list. The subscribe function is asynchronous, so it will return {:ok, ref} . Eventually a response will get delivered to the process mailbox, tagged with the reference stored in ref . It will take the form of: { { Tortoise311 , ^ client_id } , ^ ref , ^ result } Where the result can be one of :ok , or {:error, reason} . Read the documentation for Tortoise311.Connection.subscribe_sync/3 for a blocking version of this call.","ref":"Tortoise311.Connection.html#subscribe/3"},{"type":"function","title":"Tortoise311.Connection.subscribe_sync/3","doc":"Subscribe to topics and block until the server acknowledges. This is a synchronous version of the Tortoise311.Connection.subscribe/3 . In fact it calls into Tortoise311.Connection.subscribe/3 but will handle the selective receive loop, making it much easier to work with. Also, this function can be used to block a process that cannot continue before it has a subscription to the given topics. See Tortoise311.Connection.subscribe/3 for configuration options.","ref":"Tortoise311.Connection.html#subscribe_sync/3"},{"type":"function","title":"Tortoise311.Connection.subscriptions/1","doc":"Return the list of subscribed topics. Given the client_id of a running connection return its current subscriptions. This is helpful in a debugging situation.","ref":"Tortoise311.Connection.html#subscriptions/1"},{"type":"function","title":"Tortoise311.Connection.unsubscribe/3","doc":"Unsubscribe from one of more topic filters. The topic filters are given as strings. Multiple topic filters can be given at once by passing in a list of strings. Tortoise311.Connection . unsubscribe ( client_id , [ &quot;foo/bar&quot; , &quot;quux&quot; ] ) This operation is asynchronous. When the operation is done a message will be received in mailbox of the originating process.","ref":"Tortoise311.Connection.html#unsubscribe/3"},{"type":"function","title":"Tortoise311.Connection.unsubscribe_sync/3","doc":"Unsubscribe from topics and block until the server acknowledges. This is a synchronous version of Tortoise311.Connection.unsubscribe/3 . It will block until the server has send the acknowledge message. See Tortoise311.Connection.unsubscribe/3 for configuration options.","ref":"Tortoise311.Connection.html#unsubscribe_sync/3"},{"type":"module","title":"Tortoise311.Connection.Telemetry","doc":"Per-connection telemetry capture. [:tortoise311, :connection, :tx_bytes] =&gt; Number of bytes sent by the connection identified by its client id [:tortoise311, :connection, :rx_bytes] =&gt; Number of bytes received by the connection identified by its client id Meta-data: %{client_id: &lt;client_id&gt;}","ref":"Tortoise311.Connection.Telemetry.html"},{"type":"function","title":"Tortoise311.Connection.Telemetry.child_spec/1","doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Tortoise311.Connection.Telemetry.html#child_spec/1"},{"type":"function","title":"Tortoise311.Connection.Telemetry.start_link/1","doc":"","ref":"Tortoise311.Connection.Telemetry.html#start_link/1"},{"type":"module","title":"Tortoise311.Events","doc":"A PubSub exposing various system events from a Tortoise311 connection. This allows the user to integrate with custom metrics and logging solutions. Please read the documentation for Tortoise311.Events.register/2 for information on how to subscribe to events, and Tortoise311.Events.unregister/2 for how to unsubscribe.","ref":"Tortoise311.Events.html"},{"type":"function","title":"Tortoise311.Events.register/2","doc":"Subscribe to messages on the client with the client id client_id of the type type . When a message of the subscribed type is dispatched it will end up in the mailbox of the process that placed the subscription. The received message will have the format: { { Tortoise311 , client_id } , type , value } Making it possible to pattern match on multiple message types on multiple clients. The value depends on the message type. Possible message types are: :status dispatched when the connection of a client changes status. The value will be :up when the client goes online, and :down when it goes offline. :ping_response dispatched when the connection receive a response from a keep alive message. The value is the round trip time in milliseconds, and can be used to track the latency over time. Other message types exist, but unless they are mentioned in the possible message types above they should be considered for internal use only. It is possible to listen on all events for a given type by specifying :_ as the client_id .","ref":"Tortoise311.Events.html#register/2"},{"type":"function","title":"Tortoise311.Events.unregister/2","doc":"Unsubscribe from messages of type from client_id . This is the reverse of Tortoise311.Events.register/2 .","ref":"Tortoise311.Events.html#unregister/2"},{"type":"behaviour","title":"Tortoise311.Handler","doc":"User defined callback module for handling connection life cycle events. Tortoise311.Handler defines a behaviour which can be given to a Tortoise311.Connection . This allow the user to implement functionality for events happening in the life cycle of the connection, the provided callbacks are: init/1 and terminate/2 are called when the connection is started and stopped. Parts that are setup in init/1 can be torn down in terminate/2 . Notice that the connection process is not terminated if the connection to the broker is lost; Tortoise311 will attempt to reconnect, and events that should happen when the connection goes offline should be set up using the connection/3 callback. last_will/1 is called to get a last will message for a new connection. If none is provided (nil), the connection's default last will will be used. connection/3 is called when the connection is :up or :down , allowing for functionality to be run when the connection state changes. subscription/3 is called when a topic filter subscription changes status, so this callback can be used to control the life-cycle of a subscription, allowing us to implement custom behavior for when the subscription is accepted, declined, as well as unsubscribed. handle_message/3 is run when the client receive a message on one of the subscribed topic filters. Because the callback-module will run inside the connection controller process, which also handles the routing of protocol messages (such as publish acknowledge messages) it is important that the callbacks do not call functions that will block the process, especially for clients that subscribe to topics with heavy traffic. Technically it would be possible to run the callback module in a different process than the controller process, but it has been decided to keep it on the controller as we otherwise would have to copy every message evaluated by the connection to another process. It is much better to let the end-user handle the dispatching to other parts of the system while we are evaluating what to do with the process anyways with the caveats: The callbacks should not block the controller it is not possible to call the (un)subscribe and publish functions from within a callback as they will block the controller. While it is not possible to subscribe and unsubscribe in the handler process using the Tortoise311.subscribe/3 and Tortoise311.unsubscribe/3 it is possible to make changes to the subscription list via :gen_statem inspired next_actions . Next actions In some situations one would like to subscribe or unsubscribe a filter topic when a certain event happens on the client. The functions for interacting with the MQTT broker, defined on the Tortoise311 -module are for the most part blocking operations, or require the user to peek into the process mailbox to fetch the result of the operation. To allow for changes in the subscriptions one can define a set of next actions that should happen as part of the return value to the handle_message/3 , subscription/3 , and connection/3 callbacks by returning a {:ok, state, next_actions} where next_actions is a list of commands of: {:subscribe, topic_filter, qos: qos, timeout: 5000} where topic_filter is a binary containing a valid MQTT topic filter, and qos is the desired quality of service (0..2). The timeout is the amount of time in milliseconds we are willing to wait for a response to the request. {:unsubscribe, topic_filter} where topic_filter is a binary containing the name of the subscription we want to unsubscribe from. If we want to unsubscribe from the current topic when we receive a message on it we could write a handle_message/3 as follows: def handle_message ( topic , _payload , state ) do topic = Enum . join ( topic , &quot;/&quot; ) next_actions = [ { :unsubscribe , topic } ] { :ok , state , next_actions } end Note that the topic is received as a list of topic levels, and that the next actions has to be a list, even if there is only one next action; multiple actions can be given at once. Read more about this in the handle_message/3 documentation.","ref":"Tortoise311.Handler.html"},{"type":"callback","title":"Tortoise311.Handler.connection/2","doc":"Invoked when the connection status changes. status is one of :up or :down , where up means we have an open connection to the MQTT broker, and down means the connection is temporary down. The connection process will attempt to reestablish the connection. Returning {:ok, new_state} will set the state for later invocations. Returning {:ok, new_state, next_actions} , where next_actions is a list of next actions such as {:unsubscribe, &quot;foo/bar&quot;} will result in the state being returned and the next actions performed.","ref":"Tortoise311.Handler.html#c:connection/2"},{"type":"callback","title":"Tortoise311.Handler.handle_message/3","doc":"Invoked when messages are published to subscribed topics. The topic comes in the form of a list of binaries, making it possible to pattern match on the topic levels of the retrieved message, store the individual topic levels as variables and use it in the function body. Payload is a binary. MQTT 3.1.1 does not specify any format of the payload, so it has to be decoded and validated depending on the needs of the application. In an example where we are already subscribed to the topic filter room/+/temp and want to dispatch the received messages to a Temperature application we could set up our handle_message as such: def handle_message ( [ &quot;room&quot; , room , &quot;temp&quot; ] , payload , state ) do :ok = Temperature . record ( room , payload ) { :ok , state } end Notice; the handle_message/3 -callback run inside the connection controller process, so for handlers that are subscribing to topics with heavy traffic should do as little as possible in the callback handler and dispatch to other parts of the application using non-blocking calls. Returning {:ok, new_state} will reenter the loop and set the state for later invocations. Returning {:ok, new_state, next_actions} , where next_actions is a list of next actions such as {:unsubscribe, &quot;foo/bar&quot;} will reenter the loop and perform the listed actions.","ref":"Tortoise311.Handler.html#c:handle_message/3"},{"type":"callback","title":"Tortoise311.Handler.init/1","doc":"Invoked when the connection is started. args is the argument passed in from the connection configuration. Returning {:ok, state} will let the MQTT connection receive data from the MQTT broker, and the value contained in state will be used as the process state.","ref":"Tortoise311.Handler.html#c:init/1"},{"type":"callback","title":"Tortoise311.Handler.last_will/1","doc":"Invoked when a new connection is being attempted to set the last will message to one provided by the handler. If the handler returns {{:ok, nil}, state}, the pre-set last will is used.","ref":"Tortoise311.Handler.html#c:last_will/1"},{"type":"callback","title":"Tortoise311.Handler.subscription/3","doc":"Invoked when the subscription of a topic filter changes status. The status of a subscription can be one of: :up , triggered when the subscription has been accepted by the MQTT broker with the requested quality of service {:warn, [requested: req_qos, accepted: qos]} , triggered when the subscription is accepted by the MQTT broker, but with a different quality of service qos than the one requested req_qos {:error, reason} , triggered when the subscription is rejected with the reason reason such as :access_denied :down , triggered when the subscription of the given topic filter has been successfully acknowledged as unsubscribed by the MQTT broker The topic_filter is the topic filter in question, and the state is the internal state being passed through transitions. Returning {:ok, new_state} will set the state for later invocations. Returning {:ok, new_state, next_actions} , where next_actions is a list of next actions such as {:unsubscribe, &quot;foo/bar&quot;} will result in the state being returned and the next actions performed.","ref":"Tortoise311.Handler.html#c:subscription/3"},{"type":"callback","title":"Tortoise311.Handler.terminate/2","doc":"Invoked when the connection process is about to exit. If anything is setup during the init/1 callback it should get cleaned up during the terminate/2 callback.","ref":"Tortoise311.Handler.html#c:terminate/2"},{"type":"type","title":"Tortoise311.Handler.next_action/0","doc":"Action to perform before reentering the execution loop. The supported next actions are: Tell the connection process to subscribe to a topic filter Tell the connection process to unsubscribe from a topic filter More next actions might be supported in the future.","ref":"Tortoise311.Handler.html#t:next_action/0"},{"type":"type","title":"Tortoise311.Handler.t/0","doc":"Data structure describing the user defined callback handler The data structure describe the current state as well as its initial arguments and the module driving the handler. This allow Tortoise311 to restart the handler if needed be.","ref":"Tortoise311.Handler.html#t:t/0"},{"type":"module","title":"Tortoise311.Pipe","doc":"Experimental. This feature is under development. The transmitter &quot;pipe&quot;, for lack of a better word, is an opaque data type that can be given to a process. It contains amongst other things a socket. A process can obtain a transmitter pipe by issuing a pipe = Tortoise311.Pipe.new(client_id) request, which will result in a pipe in passive mode, meaning it will hold a socket it can publish messages into, but might fail, in which case it will attempt to get another socket from the transmitter. This all happens behind the scenes, it is important though that the returned pipe is used in future pipe requests, so publishing on a pipe should look like this: pipe = Tortoise311.Pipe.publish(pipe, &quot;foo/bar&quot;, &quot;bonjour !&quot;) This is all experimental, and efforts to document this better will be made when the design and implementation has stabilized.","ref":"Tortoise311.Pipe.html"},{"type":"function","title":"Tortoise311.Pipe.await/2","doc":"Await for acknowledge messages for the currently pending messages. Note that this enters a selective receive loop, so the await needs to happen before the process reaches its mailbox. It can be used in situations where we want to send a couple of messages and continue when the server has received them; This only works for messages with a Quality of Service above 0.","ref":"Tortoise311.Pipe.html#await/2"},{"type":"function","title":"Tortoise311.Pipe.new/2","doc":"Create a new publisher pipe.","ref":"Tortoise311.Pipe.html#new/2"},{"type":"function","title":"Tortoise311.Pipe.publish/4","doc":"Publish a message using a pipe.","ref":"Tortoise311.Pipe.html#publish/4"},{"type":"opaque","title":"Tortoise311.Pipe.t/0","doc":"","ref":"Tortoise311.Pipe.html#t:t/0"},{"type":"module","title":"Tortoise311.Supervisor","doc":"A dynamic supervisor that can hold Tortoise311.Connection processes A Tortoise311.Supervisor , registered under the name Tortoise311.Supervisor , is started as part of the Tortoise311.App , which will get started automatically when Tortoise311 is included as a Mix dependency. Spawning a connection on the Tortoise311.Supervisor can be done as follows: { :ok , pid } = Tortoise311.Supervisor . start_child ( client_id : T1000 , handler : { Tortoise311.Handler.Logger , [ ] } , server : { Tortoise311.Transport.Tcp , host : &#39;localhost&#39; , port : 1883 } , subscriptions : [ { &quot;foo/bar&quot; , 0 } ] ) While this is an easy way to get started one should consider the supervision strategy of the application using the Tortoise311.Connection s. Often one would like to close the connection with the application using the connection, in which case one could consider starting the Tortoise311.Connection directly in the supervisor of the application using the connection, or start a Tortoise311.Supervisor as part of ones application if it require more than one connection. See the Connection Supervision article in the project documentation for more information on connection supervision.","ref":"Tortoise311.Supervisor.html"},{"type":"function","title":"Tortoise311.Supervisor.child_spec/1","doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Tortoise311.Supervisor.html#child_spec/1"},{"type":"function","title":"Tortoise311.Supervisor.start_child/2","doc":"Start a connection as a child of the Tortoise311.Supervisor . supervisor is the name of the supervisor the child should be started on, and it defaults to Tortoise311.Supervisor .","ref":"Tortoise311.Supervisor.html#start_child/2"},{"type":"function","title":"Tortoise311.Supervisor.start_link/1","doc":"Start a dynamic supervisor that can hold connection processes. The :name option can also be given in order to register a supervisor name, the supported values are described in the &quot;Name registration&quot; section in the GenServer module docs.","ref":"Tortoise311.Supervisor.html#start_link/1"},{"type":"behaviour","title":"Tortoise311.Transport","doc":"Abstraction for working with network connections; this is done to normalize the :ssl and :gen_tcp modules, so they get a similar interface. This work has been heavily inspired by the Ranch project by NineNines.","ref":"Tortoise311.Transport.html"},{"type":"callback","title":"Tortoise311.Transport.accept/2","doc":"","ref":"Tortoise311.Transport.html#c:accept/2"},{"type":"callback","title":"Tortoise311.Transport.accept_ack/2","doc":"","ref":"Tortoise311.Transport.html#c:accept_ack/2"},{"type":"callback","title":"Tortoise311.Transport.close/1","doc":"","ref":"Tortoise311.Transport.html#c:close/1"},{"type":"callback","title":"Tortoise311.Transport.connect/4","doc":"","ref":"Tortoise311.Transport.html#c:connect/4"},{"type":"callback","title":"Tortoise311.Transport.controlling_process/2","doc":"","ref":"Tortoise311.Transport.html#c:controlling_process/2"},{"type":"callback","title":"Tortoise311.Transport.getopts/2","doc":"","ref":"Tortoise311.Transport.html#c:getopts/2"},{"type":"callback","title":"Tortoise311.Transport.getstat/1","doc":"","ref":"Tortoise311.Transport.html#c:getstat/1"},{"type":"callback","title":"Tortoise311.Transport.getstat/2","doc":"","ref":"Tortoise311.Transport.html#c:getstat/2"},{"type":"callback","title":"Tortoise311.Transport.listen/1","doc":"","ref":"Tortoise311.Transport.html#c:listen/1"},{"type":"function","title":"Tortoise311.Transport.new/1","doc":"Create a new Transport specification used by the Connection process to log on to the MQTT server. This allow us to filter the options passed to the connection type, and guide the user to connect to the individual transport type.","ref":"Tortoise311.Transport.html#new/1"},{"type":"callback","title":"Tortoise311.Transport.new/1","doc":"","ref":"Tortoise311.Transport.html#c:new/1"},{"type":"callback","title":"Tortoise311.Transport.peername/1","doc":"","ref":"Tortoise311.Transport.html#c:peername/1"},{"type":"callback","title":"Tortoise311.Transport.recv/3","doc":"","ref":"Tortoise311.Transport.html#c:recv/3"},{"type":"callback","title":"Tortoise311.Transport.send/2","doc":"","ref":"Tortoise311.Transport.html#c:send/2"},{"type":"callback","title":"Tortoise311.Transport.setopts/2","doc":"","ref":"Tortoise311.Transport.html#c:setopts/2"},{"type":"callback","title":"Tortoise311.Transport.shutdown/2","doc":"","ref":"Tortoise311.Transport.html#c:shutdown/2"},{"type":"callback","title":"Tortoise311.Transport.sockname/1","doc":"","ref":"Tortoise311.Transport.html#c:sockname/1"},{"type":"type","title":"Tortoise311.Transport.opts/0","doc":"","ref":"Tortoise311.Transport.html#t:opts/0"},{"type":"type","title":"Tortoise311.Transport.socket/0","doc":"","ref":"Tortoise311.Transport.html#t:socket/0"},{"type":"type","title":"Tortoise311.Transport.stats/0","doc":"","ref":"Tortoise311.Transport.html#t:stats/0"},{"type":"opaque","title":"Tortoise311.Transport.t/0","doc":"","ref":"Tortoise311.Transport.html#t:t/0"},{"type":"extras","title":"Introduction","doc":"Tortoise311 is a MQTT client application written in Elixir for use in Elixir applications. Its philosophy is to separate the MQTT semantics from the Elixir semantics, attempting to hide the MQTT protocol details from the end-user while still being flexible enough to be able to do just about anything one would like to do with the MQTT protocol.","ref":"introduction.html"},{"type":"extras","title":"Introduction - MQTT overview","doc":"MQTT is a lightweight Publish/Subscribe (PubSub) protocol, designed to operate on high-latency networks or unreliable networks. It is a server/client architecture, so a MQTT server (often referred to as a broker) is required for the clients to connect to. All communication happens via the broker, clients cannot see each other, and the capabilities of the MQTT broker can vary depending on its implementation and its configuration. Being a PubSub protocol the clients can connect to the broker and subscribe to topics, by specifying topic filters, and post messages to topics. Depending on the configuration of the broker the topics can be free-form (any valid topic filter accepted), or restricted to a handful of topics. A topic consist of one or more topic levels , separated by forward slashes such as: rooms / living - room / temp rooms / hall / temp rooms / kitchen / temp &quot;Topic filters&quot; allow the client to glob on topics, and support wildcards for a single level using a plus sign + ; or multiple levels using a hash sign # , which may only be at the last spot of the topic filter. rooms/+/temp would subscribe to messages posted to all the topics in the previous example, and rooms/# would subscribe to every message posted to all topics starting with rooms . The wildcard patterns can be mixed, so +/+/# would be a valid topic. When publishing a message we cannot use the wildcard functionality, that is strictly for subscribing; Publishing messages needs to be addressed to a specific topic, not a topic filter . The protocol support three kinds of quality of service (QoS) for both subscriptions and publishing. The three kinds are defined as the following: QoS=0 is a fire and forget. The sender has no guarantee that the receiver will receive the message. QoS=1 where the receiver will respond with an acknowledge message, at which point the sender will stop retransmitting the message. It is possible that a retransmission will arrive after the acknowledge message has been sent, in which case the retransmission will be treated as a new message. In other words, this implements at least once delivery . QoS=2 implements a series of acknowledge and release messages ensuring a message will only get delivered once to the receiver, referred to as exactly once delivery . The higher quality of service specified the more expensive a publish or a subscription will be bandwidth wise, as a higher QoS will require more protocol messages to fulfill network exchange. That is why some broker administrators choose to configure their broker to refuse a higher quality of service for some topics, or some subscriptions. In the case of subscriptions it can happen that a lower QoS than the one requested is accepted. Note : Placing a subscription with QoS=2 does not mean that the client will receive messages with QoS=2, it will receive all messages published to that topic regardless of QoS; the message exchange between the broker and the client will use the specified QoS in the subscription. When a message is published to the broker, the client sending the message can specify that the broker should retain the message; if so the broker will keep the message and dispatch it to clients when they subscribe to the topic. It is also possible for a client to specify a last will message when connecting to the broker. The broker will keep this message and dispatch it to the specified topic when the client disconnects abruptly from the broker.","ref":"introduction.html#mqtt-overview"},{"type":"extras","title":"Introduction - The Tortoise311 MQTT Client","doc":"Tortoise311 aims to hide the MQTT semantics from the user, and expose an interface that should be familiar to an Elixir developer. This means the message exchanges needed to complete a publish or retrieval of a message on a subscription with a QoS&gt;0 should be handled in the background and the details of the protocol should not bleed through to the user. While the goal of the project is to hide the details of the MQTT protocol from the user, another goal is not to restrict the user in any way. This means Tortoise311 will attempt to map Elixir semantics to MQTT semantics, which should be possible because both systems deal with message passing: A publish with QoS=0 works like a cast; send and forget. A publish with QoS&gt;0 results in a message exchange between the sender and the receiver; to keep track of these messages the MQTT protocol specifies that these messages should have an identifier assigned, a random 16-bit number. The user of Tortoise311 will never see this random 16-bit number, but instead an Erlang reference will get returned from the publish function, which makes it possible to await and pattern match for the return. Furthermore, when it comes to publishing messages; the client should make sure that messages with a QoS&gt;0 are delivered, so if the client is offline it should store these messages and send them when the client is back online. A callback behaviour is specified allowing the user to implement custom behavior for when the client connects; disconnects; accepts a subscription; acknowledges an unsubscribe; and receives a message on one of the subscribed topics. If the design of Tortoise311 is a hindrance to creating something with MQTT (within reason) it should be considered a bug, and it should be discussed how we should map that behaviour to Elixir semantics.","ref":"introduction.html#the-tortoise311-mqtt-client"},{"type":"extras","title":"Introduction - Summary","doc":"MQTT is a protocol implementing a PubSub allowing one or multiple clients to subscribe to topics, using topic filters, and clients to publish messages to topics. Tortoise311 is an MQTT client that aims to map MQTT semantics to Elixir so sending and receiving messages using an MQTT broker feels like any other kind of message passing, and awaiting results feels natural to the Elixir ecosystem.","ref":"introduction.html#summary"},{"type":"extras","title":"Connecting to a MQTT Broker","doc":"Tortoise311 is capable of connecting to any MQTT broker that implements the 3.1.1 version of the MQTT protocol. It does so by taking a connection specification, and with it will do its best to connect to the broker and keeping the connection open. A minimal connection specification looks like this: { ok , _pid } = Tortoise311.Connection . start_link ( client_id : HelloWorld , server : { Tortoise311.Transport.Tcp , host : &quot;localhost&quot; , port : 1883 } , handler : { Tortoise311.Handler.Logger , [ ] } ) This will establish a TCP connection to a broker running on localhost port 1883 . The connection takes a module that implements the Tortoise311.Handler behaviour; In this case the Tortoise311.Handler.Logger callback module, which will print a log statement on events happening during the connection life cycle. Furthermore, we specify that the client_id of the connection is HelloWorld . The client id can later be used to interact with the connection, such as publishing messages and subscribing to topics. Notice that this example expects a server configured to allow anonymous connections without SSL. Not all MQTT brokers are configured the same, so depending on the server more configuration options might be needed for a successful connection. This document aims to give an overview.","ref":"connecting_to_a_mqtt_broker.html"},{"type":"extras","title":"Connecting to a MQTT Broker - Network Transport","doc":"Tortoise311 has an abstraction for the network transport, and comes with two official implementations included in Tortoise311 itself: Tortoise311.Transport.Tcp used to connect to a broker via TCP. While this transport is the simplest to use it is also the least secure, and should only be used on trusted networks. It is based on :gen_tcp found in the Erlang/OTP distribution. Tortoise311.Transport.SSL used to create a secure connection via secure socket layer. This option takes a bit more work to setup, but it will prevent people from eavesdropping on the data being sent between the client and the broker. The transports are given with the server field in the connection specification as a tuple, containing the transport type and an options list specific to the given transport. Tortoise311.Transport.Tcp takes two options; the host name as a string, such as &quot;localhost&quot; or a four-tuple describing an IP-network address, such as {127, 0, 0, 1} . An example where the TCP transport is used can be seen in the introduction to this article. The Tortoise311.Transport.SSL is a bit more versatile in its configuration options. The most important additional options are: cacertfile needs to point to a file with trusted CA certificates, which is necessary for server certificate verification. For instance, when using the certifi package, pass cacertfile: :certifi.cacertfile() in the options. It is also possible to pass a list of binary (DER encoded) root CA certificates using the cacerts option instead. certfile and keyfile are needed if the server authenticates clients using a client certificate. The cert option can be used to pass the client certificate as DER binary, and the key option can be used to pass the key using Erlang's ssl format (e.g. {:RSAPrivateKey, der_binary_key} ). verify defaults to :verify_peer , enabling server certificate verification. To override, include verify: :verify_none in the server options. In that case there is no need to define CA certificates, but an attacker could intercept the connection without detection! Note that any paths must be specified as charlists. The implementation is based on the :ssl module from the Erlang distribution, so be sure to check the documentation for the :ssl module for detailed information on the possible configuration options. Information on creating a custom transport can be found in the Tortoise311.Transport module, but for most cases the TCP and SSL modules should suffice.","ref":"connecting_to_a_mqtt_broker.html#network-transport"},{"type":"extras","title":"Connecting to a MQTT Broker - Connection Handler","doc":"A handful of events are possible during a client life cycle. Tortoise311 aims to expose the interesting events as callback functions, defined in the Tortoise311.Handler behaviour, making it possible to implement custom behavior for the client. The exposed events are: The client is initialized, or terminated allowing for initialization and tear down of subsystems A connection to the server is established The connection to the server is dropped The subscription status of a given topic filter is changed A message is received on one of the subscribed topic filters Read more about defining custom behavior for a connection in the documentation for the Tortoise311.Handler module.","ref":"connecting_to_a_mqtt_broker.html#connection-handler"},{"type":"extras","title":"Connecting to a MQTT Broker - The client_id","doc":"In MQTT, the clients announce themselves to the broker with what is referred to as a client id . Two clients cannot share the same client id on a broker, and depending on the implementation (or configuration) the server will either kick the first client out, or deny the new client if it specifies a client id that is already in use. The protocol specifies that a valid client id is between 1 and 23 UTF-8 encoded bytes in length, but some server configurations may allow for longer ids; thus tortoise will allow for client identifiers longer than 23 bytes but some MQTT brokers might reject the connection. Allowed values are a string or an atom. If an atom is specified it will be converted to a string when the connection message is sent on the wire, but it will be possible to refer to the connection using the atom, which can be more convenient. Notice that the client id can easily reach the 23 bytes when converted from an atom because atoms starting with an uppercase letter will be prefixed with Elixir. ; therefore MyClientId will be 17 bytes instead of the 10 one could expect. iex(1)&gt; client_id = Atom . to_string ( MyClientId ) &quot;Elixir.MyClientId&quot; iex(2)&gt; byte_size ( client_id ) 17 The specified client identifier is used to identify a connection when publishing messages, subscribing to topics, or otherwise interacting with the named connection. { ok , _pid } = Tortoise311.Connection . start_link ( client_id : MyClient , server : { Tortoise311.Transport.Tcp , host : &quot;localhost&quot; , port : 1883 } , handler : { Tortoise311.Handler.Logger , [ ] } ) Tortoise311 . publish ( MyClient , &quot;foo/bar&quot; , &quot;hello&quot; ) Note : Though the MQTT 3.1.1 protocol allows for a zero-byte client id — in which case the server should assign a random client_id for the connection — a client id is enforced in Tortoise311. This is done so the connection has an identifier that can be used when interacting with the connection.","ref":"connecting_to_a_mqtt_broker.html#the-client_id"},{"type":"extras","title":"Connecting to a MQTT Broker - User Name and Password","doc":"Some brokers are configured to require some basic authentication, which will determine whether a user is allowed to subscribe or publish to a given topic, and some set limitations to what quality of service a particular user, or group of users, are allowed to subscribe or publish with. To specify a user name and password for a connection the aptly named user_name and password connection configuration fields come in handy. Both of them take UTF-8 encoded strings, or nil as their value, in which case an anonymous connection is attempted. Depending on the broker configuration it is allowed to specify a user name and omit the password, but the user name has to be specified if a password is specified. Both default to nil if left blank.","ref":"connecting_to_a_mqtt_broker.html#user-name-and-password"},{"type":"extras","title":"Connecting to a MQTT Broker - The keep alive interval","doc":"When connected, an MQTT client should ping the server on a set interval to let the broker know that it is still alive. The keep alive value is given as an integer, describing time in seconds between keep alive messages, and should be set depending on factors such as power consumption, network bandwidth, etc. Per default Tortoise311 will send a keep alive message every 60 seconds, which is a reasonable value for most installations. The allowed maximum value is 65_535 , which is 18 hours, 12 minutes, and 15 seconds; most would consider this a bit too extreme, and some brokers might reject connections specifying keep_alive interval that is too high. Some brokers allow disabling the keep alive interval by setting it to zero, so Tortoise311 allows for a keep_alive specified as 0 . Note that the broker can still choose to disconnect a given client due to inactivity. When keep_alive is disabled, the broker implementation will decide its own measure of inactivity. So, to avoid unspecified behavior, it is advised to use a keep alive value.","ref":"connecting_to_a_mqtt_broker.html#the-keep-alive-interval"},{"type":"extras","title":"Connecting to a MQTT Broker - Last will message","doc":"It is possible to specify a message which should be dispatched by the broker if the client is abruptly disconnected from the broker. This message is known as the last will message, and allows for other connected clients to act on other clients leaving the broker. The (default) last will message is specified as part of the connection, and for Tortoise311 it is possible to configure a last will message by passing in a Tortoise311.Package.Publish struct to the will connection configuration field. Note that the Tortoise311 handler can provide a new last will message for each new connection. If the handler chooses not to, the connection's default message will be used. { :ok , pid } = Tortoise311.Connection . start_link ( client_id : William , server : { Tortoise311.Transport.Tcp , host : &#39;localhost&#39; , port : 1883 } , handler : { Tortoise311.Handler.Logger , [ ] } , will : % Tortoise311.Package.Publish { topic : &quot;foo/bar&quot; , payload : &quot;goodbye&quot; } ) If we have another client connected to the broker, subscribing to foo/bar , we should now receive a message containing the message goodbye on that topic, should the client called William disconnect abruptly from the broker. We can simulate this by terminating the pid using Process.exit(pid, :ouch) .","ref":"connecting_to_a_mqtt_broker.html#last-will-message"},{"type":"extras","title":"Connection Supervision","doc":"An important aspect of building an Elixir application is setting up a supervision structure that ensures the application will continue to work if parts of the system should reach an erroneous state and need to get restarted into a known working state. To do this one needs to group the processes the application consist of in a manner such that processes belonging together will start and terminate together. Tortoise311 offers multiple ways of supervising one or multiple connections; by using the provided dynamic Tortoise311.Supervisor or starting a dynamic supervisor belonging to the application using the connections; or by starting the connections needed directly in an application supervisor. This document will describe the ways of supervision, and give an overview for when to use a given supervision strategy.","ref":"connection_supervision.html"},{"type":"extras","title":"Connection Supervision - Linked Connection","doc":"A connection can be started and linked to the current process by using the Tortoise311.Connection.start_link/1 function. Tortoise311.Connection . start_link ( client_id : HeartOfGold , server : { Tortoise311.Transport.Tcp , host : &#39;localhost&#39; , port : 1883 } , handler : { Tortoise311.Handler.Logger , [ ] } ) As with any other linked process both process will terminate if either terminate, as described in the Process.link/1 documentation. This means that any stored state in the process that owns the MQTT connection will disappear with the process if the connection process terminates. Therefore it is not recommended to link a connection process like this outside of experimenting in IEx , but instead to run it inside of a supervisor process. When properly supervised connections terminate, the crash will be contained, allowing the other processes to keep their state.","ref":"connection_supervision.html#linked-connection"},{"type":"extras","title":"Connection Supervision - Supervising a connection","doc":"The Tortoise311.Connection module provides a child_spec/1 which makes it easier to start a Tortoise311.Connection as part of a supervisor by simply passing a {Tortoise311.Connection, connection_specification} to the supervisor child list. defmodule MyApp.Supervisor do use Supervisor def start_link ( opts ) do Supervisor . start_link ( __MODULE__ , opts , name : __MODULE__ ) end @impl true def init ( _opts ) do children = [ { Tortoise311.Connection , [ client_id : WombatTaskForce , server : { Tortoise311.Transport.Tcp , host : &#39;localhost&#39; , port : 1883 } , handler : { Tortoise311.Handler.Logger , [ ] } ] } ] Supervisor . init ( children , strategy : :one_for_one ) end end The great thing about this approach is that the connection can live in the same supervision tree as the rest of the application that depends on that connection. The connection is started, restarted, and stopped with the application as a whole, ensuring the connection is closed with the processes that depend on it. Be sure to set a reasonable connection strategy for the supervisor. Refer to the Supervisor documentation for more information on usage and configuration.","ref":"connection_supervision.html#supervising-a-connection"},{"type":"extras","title":"Connection Supervision - The Tortoise311.Supervisor","doc":"When Tortoise311 is included as a dependency in the mix.exs file of an application, Tortoise311 will automatically get started alongside the application. During the application start up a dynamic supervisor will spawn and register itself under the name Tortoise311.Supervisor . This can be used to start supervised connections that will get restarted if they are terminated with an abnormal reason. To start a connection on the Tortoise311.Supervisor one can use the Tortoise311.Supervisor.start_child/2 function, which defaults to using the dynamic supervisor registered under the name Tortoise311.Supervisor . Tortoise311.Supervisor . start_child ( client_id : &quot;heart-of-gold&quot; , handler : { Tortoise311.Handler.Logger , [ ] } , server : { Tortoise311.Transport.Tcp , host : &#39;localhost&#39; , port : 1883 } ) This is an easy and convenient way of getting started, as everything needed to supervise a connection is there when the Tortoise311 application has been initialized. One downside is that while the children are supervised they are not grouped with the application that uses the connections; they are grouped with the Tortoise311 application. To mitigate this, a Tortoise311.Supervisor.child_spec/1 function is available, which can be used to start the Tortoise311.Supervisor as part of another supervisor. defmodule MyApp.Supervisor do use Supervisor def start_link ( opts ) do Supervisor . start_link ( __MODULE__ , opts , name : __MODULE__ ) end @impl true def init ( _opts ) do children = [ { Tortoise311.Supervisor , [ name : MyApp.Connection.Supervisor , strategy : :one_for_one ] } ] Supervisor . init ( children , strategy : :one_for_one ) end end Connections can now, dynamically, be attached to the supervised Tortoise311.Supervisor by calling the Tortoise311.Supervisor.start_child/2 function with the name that was given to the supervisor, in this case MyApp.Connection.Supervisor . Tortoise311.Supervisor . start_child ( MyApp.Connection.Supervisor , client_id : SmartHose , server : { Tortoise311.Transport.Tcp , host : &#39;localhost&#39; , port : 1883 } , handler : { Tortoise311.Handler.Logger , [ ] } ) This is the best way of supervising a dynamic set of connections, but might be overkill if only one static connection is needed for the application.","ref":"connection_supervision.html#the-tortoise311-supervisor"},{"type":"extras","title":"Connection Supervision - Summary","doc":"Tortoise311 makes it possible to spawn connections and supervise them, and it is always a best practice to supervise a connection to ensure it remains up. Different approaches can be taken depending on the situation: If a fixed number of connections are needed, the recommended way is to attach them directly to a supervision tree, along with the processes that depend on said connections using the Tortoise311.Connection.child_spec/1 function. If a dynamic set of connections is needed, the recommended way is to spawn a named Tortoise311.Supervisor as part of a supervisor, which holds the processes that depend on the connections, and spawn the connections on the dynamic supervisor. Supervising the connections along the processes that rely on them ensures that the application can be started and stopped as a whole, and makes it possible to recover from a faulty state.","ref":"connection_supervision.html#summary"},{"type":"extras","title":"Publishing Messages","doc":"Tortoise311 provides a couple of methods for publishing messages to an MQTT broker. They vary in the amount of setup needed to setup a publish, and they differ in the way connection drops are handled. This document will describe the different supported ways of publishing messages, and should serve as a guide for when to choose a particular strategy and its setup. All publish methods require an open connection.","ref":"publishing_messages.html"},{"type":"extras","title":"Publishing Messages - Tortoise311.publish/4","doc":"The simplest way of publishing a message is to use the publish function found on the main Tortoise311 module. Given the identifier ( client_id ) of an open connection, and a topic, such as foo/bar , a publish is as simple as Tortoise311.publish(MyClientId, &quot;foo/bar&quot;) , which will send a message to the topic foo/bar with an empty payload ( nil ). A message payload can be specified by passing data as the third argument. Tortoise311 . publish ( MyClientId , &quot;foo/bar&quot; , &quot;hello, world !&quot; ) We can pass in a fourth argument: an option list allowing us to specify: qos (default: 0 ) the requested quality of service retain (default: false ) specifying if the message should be retained on the server for the given topic Here we will publish a retained message with QoS=1 to baz . Tortoise311 . publish ( MyClientId , &quot;baz&quot; , &quot;Hi !&quot; , [ qos : 1 , retain : true ] ) If we want to pass options to a message without a payload we should specify the payload as nil Tortoise311 . publish ( MyClientId , &quot;baz&quot; , nil , [ qos : 2 ] ) If we were to pass in an empty string ( &quot;&quot; , or &lt;&lt;&gt;&gt; ) a zero length binary would be sent instead, which is different from an empty payload. Notice that the semantics of the returned value change when the QoS is bigger than 0 . For messages published with QoS=0 the client will respond with :ok . iex(1)&gt; Tortoise311 . publish ( MyClientId , &quot;baz&quot; , nil , [ qos : 0 ] ) :ok A message with QoS=0 has no quality of service, which in MQTT terms means that we have no guarantee that the message was delivered. If the message is published with QoS=1 or QoS=2 we will get a reference as the return value. This reference allows us to enter a selective receive where we can get the result of the publish; in an iex context we can simply use the flush/0 command to see the result: iex(2)&gt; Tortoise311 . publish ( MyClientId , &quot;baz&quot; , nil , [ qos : 1 ] ) { :ok , # Reference &lt; 0.1924528969 . 904134659.102547 &gt; } iex(3)&gt; flush ( ) { { Tortoise311 , C } , # Reference &lt; 0.1924528969 . 904134659.102547 &gt; , :ok } :ok As observed we can expect a message containing a three-tuple. The first element contains an identifier for the tortoise connection {Tortoise311, *client identifier*} , then the reference created by the publish command, and finally the result of the operation; in this case :ok . This structure should allow for some flexible pattern matches in handle_info/2 when Tortoise311 is used in the context of a GenServer .","ref":"publishing_messages.html#tortoise311-publish-4"}]